#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#ifndef PORT
#define PORT 10000
#endif

#define BUFSZ 128
#define MAX_ATTEMPTS 2

char password[BUFSZ];

int readpw(int fd, char *pw)
{
	int n, m;

	n = BUFSZ;
	while (n > 0) {
		m = read(fd, pw, n);
		if (m <= 0)
			return -1;
		n -= m;
		pw += m;
	}
	return 0;
}

int login()
{
	char buf[BUFSZ];
	char errmsg[BUFSZ] = "wrong! remaining attempts: %d\n";
	int avail, errors, good;
	char *pw1, *pw2;

	avail = MAX_ATTEMPTS;
	pw1 = buf;
retry:
	if (readpw(0, pw1) < 0)
		return -1;
	pw2 = password;
	errors = 0;
	good = 0;
	// we check the whole BUFSZ, to prevent timing attacks
	while (pw2 != password + BUFSZ) {
		if (*pw1++ != *pw2++) {
			errors++;
		} else {
			good++;
		}
	}
	if (errors) {
		avail--;
		printf(errmsg, avail);
		if (avail > 0)
			goto retry;
		return -1;
	}
	return 0;
}

void readflag()
{
	int fd;

	fd = open("flag.txt", O_RDONLY);
	if (fd < 0) {
		perror("flag.txt");
		exit(1);
	}
	if (readpw(fd, password) < 0) {
		fprintf(stderr, "invalid configuration\n");
		exit(1);
	}
	close(fd);
}

void child()
{
	printf("Please enter your password\n");
	if (login() < 0) {
		printf("sorry\n");
		exit(1);
	}
	printf("OK\n");
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;

	setvbuf(stdin, NULL, _IONBF, 0);
	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	readflag();

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, SIG_IGN);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			dup(con);
			close(1);
			dup(con);
			close(2);
			dup(con);
			close(con);
			child();
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;
}
